/**************************************************************************
    Lightspark, a free flash player implementation

    Copyright (C) 2024  mr b0nk 500 (b0nk@b0nk.xyz)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************/

/* This file was generated by forward-declare-gen.sh. - DO NOT EDIT */

#ifndef FORWARDS_PARSING_TAGS_H
#define FORWARDS_PARSING_TAGS_H 1

namespace lightspark
{

/* forward declarations */
class Tag;
class EndTag;
class DisplayListTag;
class DictionaryTag;
class ControlTag;
class ActionTag;
class AVM1ActionTag;
class AVM1InitActionTag;
class DefineShapeTag;
class DefineShape2Tag;
class DefineShape3Tag;
class DefineShape4Tag;
class DefineMorphShapeTag;
class DefineMorphShape2Tag;
class DefineEditTextTag;
class DefineSoundTag;
class StartSoundTag;
class SoundStreamHeadTag;
class SoundStreamBlockTag;
class ShowFrameTag;
class RemoveObject2Tag;
class PlaceObject2Tag;
class PlaceObject3Tag;
class FrameLabelTag;
class SetBackgroundColorTag;
class DefineButtonSoundTag;
class DefineButtonTag;
class KERNINGRECORD;
class DefineBinaryDataTag;
class FontTag;
class DefineFontTag;
class DefineFontInfoTag;
class DefineFont2Tag;
class DefineFont3Tag;
class DefineFont4Tag;
class DefineTextTag;
class DefineText2Tag;
class DefineSpriteTag;
class ProtectTag;
class BitmapTag;
class JPEGTablesTag;
class DefineBitsLosslessTag;
class DefineBitsTag;
class DefineBitsJPEG2Tag;
class DefineBitsJPEG3Tag;
class DefineScalingGridTag;
class AdditionalDataTag;
class UnimplementedTag;
class DefineSceneAndFrameLabelDataTag;
class DefineFontNameTag;
class DefineFontAlignZonesTag;
class DefineVideoStreamTag;
class VideoFrameTag;
class MetadataTag;
class CSMTextSettingsTag;
class ScriptLimitsTag;
class ProductInfoTag;
class FileAttributesTag;
class EnableDebuggerTag;
class EnableDebugger2Tag;
class DebugIDTag;
class ExportAssetsTag;
class NameCharacterTag;
class TagFactory;
class DoABCTag;
class DoABCDefineTag;
class SymbolClassTag;

};
#endif /* FORWARDS_PARSING_TAGS_H */
