/**************************************************************************
    Lightspark, a free flash player implementation

    Copyright (C) 2024  mr b0nk 500 (b0nk@b0nk.xyz)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************/

/* This file was generated by forward-declare-gen.sh. - DO NOT EDIT */

#ifndef FORWARDS_SCRIPTING_FLASH_SYSTEM_FLASHSYSTEM_H
#define FORWARDS_SCRIPTING_FLASH_SYSTEM_FLASHSYSTEM_H 1

namespace lightspark
{

/* forward declarations */
class Capabilities;
class ApplicationDomain;
class LoaderContext;
class SecurityDomain;
class Security;
class System;
class ASWorker;
class WorkerDomain;
class WorkerState;
class ImageDecodingPolicy;
class IMEConversionMode;

};
#endif /* FORWARDS_SCRIPTING_FLASH_SYSTEM_FLASHSYSTEM_H */
