/**************************************************************************
    Lightspark, a free flash player implementation

    Copyright (C) 2024  mr b0nk 500 (b0nk@b0nk.xyz)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************/

/* This file was generated by forward-declare-gen.sh. - DO NOT EDIT */

#ifndef FORWARDS_SCRIPTING_FLASH_EVENTS_FLASHEVENTS_H
#define FORWARDS_SCRIPTING_FLASH_EVENTS_FLASHEVENTS_H 1

namespace lightspark
{

/* forward declarations */
class Event;
class WaitableEvent;
class EventPhase;
class KeyboardEvent;
class FocusEvent;
class FullScreenEvent;
class NetStatusEvent;
class HTTPStatusEvent;
class TextEvent;
class ErrorEvent;
class IOErrorEvent;
class SecurityErrorEvent;
class AsyncErrorEvent;
class UncaughtErrorEvent;
class ProgressEvent;
class TimerEvent;
class MouseEvent;
class NativeDragEvent;
class NativeWindowBoundsEvent;
class InvokeEvent;
class listener;
class IEventDispatcher;
class EventDispatcher;
class StatusEvent;
class DataEvent;
class BindClassEvent;
class ShutdownEvent;
class FunctionEvent;
class FunctionAsyncEvent;
class ExternalCallEvent;
class ABCContextInitEvent;
class AVM1InitActionEvent;
class InitFrameEvent;
class ExecuteFrameScriptEvent;
class AdvanceFrameEvent;
class SetLoaderContentEvent;
class RootConstructedEvent;
class LocalConnectionEvent;
class IdleEvent;
class GetMouseTargetEvent;
class FlushEventBufferEvent;
class FlushInvalidationQueueEvent;
class ParseRPCMessageEvent;
class TextInputEvent;
class DRMErrorEvent;
class DRMStatusEvent;
class VideoEvent;
class StageVideoEvent;
class StageVideoAvailabilityEvent;
class ContextMenuEvent;
class TouchEvent;
class GestureEvent;
class PressAndTapGestureEvent;
class TransformGestureEvent;
class UncaughtErrorEvents;
class SampleDataEvent;
class ThrottleEvent;
class ThrottleType;
class GameInputEvent;

};
#endif /* FORWARDS_SCRIPTING_FLASH_EVENTS_FLASHEVENTS_H */
