/**************************************************************************
    Lightspark, a free flash player implementation

    Copyright (C) 2024  mr b0nk 500 (b0nk@b0nk.xyz)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**************************************************************************/

/* This file was generated by forward-declare-gen.sh. - DO NOT EDIT */

#ifndef FORWARDS_SWFTYPES_H
#define FORWARDS_SWFTYPES_H 1

namespace lightspark
{

/* forward declarations */
class QName;
class UI8;
class UI16_SWF;
class UI16_FLV;
class SI16_SWF;
class SI16_FLV;
class UI24_SWF;
class UI24_FLV;
class SI24_SWF;
class SI24_FLV;
class UI32_SWF;
class UI32_FLV;
class u32;
class STRING;
struct nsNameAndKindImpl;
struct nsNameAndKind;
struct multiname;
class FLOAT;
class DOUBLE;
class FIXED;
class FIXED8;
class RECORDHEADER;
class RGB;
class RGBA;
class BitStream;
class FB;
class UB;
class SB;
class RECT;
class MATRIX;
class GRADRECORD;
class GRADIENT;
class FOCALGRADIENT;
class FILLSTYLE;
class MORPHFILLSTYLE;
class LINESTYLE;
class LINESTYLE2;
class MORPHLINESTYLE;
class MORPHLINESTYLE2;
class LINESTYLEARRAY;
class MORPHLINESTYLEARRAY;
class FILLSTYLEARRAY;
class MORPHFILLSTYLEARRAY;
class SHAPERECORD;
class GLYPHENTRY;
class TEXTRECORD;
class SHAPE;
class SHAPEWITHSTYLE;
class CXFORMWITHALPHA;
class CXFORM;
class DROPSHADOWFILTER;
class BLURFILTER;
class GLOWFILTER;
class BEVELFILTER;
class GRADIENTGLOWFILTER;
class CONVOLUTIONFILTER;
class COLORMATRIXFILTER;
class GRADIENTBEVELFILTER;
class FILTER;
class FILTERLIST;
class BUTTONRECORD;
class CLIPEVENTFLAGS;
class CLIPACTIONRECORD;
class CLIPACTIONS;
class SOUNDENVELOPE;
class SOUNDINFO;
class RunState;
class ACTIONRECORD;
class BUTTONCONDACTION;

};
#endif /* FORWARDS_SWFTYPES_H */
